@@ -3,38 +3,46 @@
 #
 # This module is part of GitPython and is released under
 # the BSD License: http://www.opensource.org/licenses/bsd-license.php
+
 import re
+from git.cmd import handle_process_output
+from git.compat import defenc
+from git.util import finalize_process, hex_to_bin
 
-from gitdb.util import hex_to_bin
-
-from .compat import binary_type
 from .objects.blob import Blob
 from .objects.util import mode_str_to_int
 
-from git.compat import (
-    defenc,
-    PY3
-)
+
+# typing ------------------------------------------------------------------
+
+from typing import Any, Iterator, List, Match, Optional, Tuple, Type, Union, TYPE_CHECKING
+from typing_extensions import Final, Literal
+from git.types import TBD
+
+if TYPE_CHECKING:
+    from .objects.tree import Tree
+    from git.repo.base import Repo
+
+Lit_change_type = Literal['A', 'D', 'M', 'R', 'T']
+
+# ------------------------------------------------------------------------
 
 __all__ = ('Diffable', 'DiffIndex', 'Diff', 'NULL_TREE')
 
 # Special object to compare against the empty tree in diffs
-NULL_TREE = object()
+NULL_TREE = object()  # type: Final[object]
 
 _octal_byte_re = re.compile(b'\\\\([0-9]{3})')
 
 
-def _octal_repl(matchobj):
+def _octal_repl(matchobj: Match) -> bytes:
     value = matchobj.group(1)
     value = int(value, 8)
-    if PY3:
-        value = bytes(bytearray((value,)))
-    else:
-        value = chr(value)
+    value = bytes(bytearray((value,)))
     return value
 
 
-def decode_path(path, has_ab_prefix=True):
+def decode_path(path: bytes, has_ab_prefix: bool = True) -> Optional[bytes]:
     if path == b'/dev/null':
         return None
 
@@ -60,13 +68,13 @@ class Diffable(object):
     :note:
         Subclasses require a repo member as it is the case for Object instances, for practical
         reasons we do not derive from Object."""
-    __slots__ = tuple()
+    __slots__ = ()
 
     # standin indicating you want to diff against the index
     class Index(object):
         pass
 
-    def _process_diff_args(self, args):
+    def _process_diff_args(self, args: List[Union[str, 'Diffable', object]]) -> List[Union[str, 'Diffable', object]]:
         """
         :return:
             possibly altered version of the given args list.
@@ -74,7 +82,9 @@ class Diffable(object):
             Subclasses can use it to alter the behaviour of the superclass"""
         return args
 
-    def diff(self, other=Index, paths=None, create_patch=False, **kwargs):
+    def diff(self, other: Union[Type[Index], Type['Tree'], object, None, str] = Index,
+             paths: Union[str, List[str], Tuple[str, ...], None] = None,
+             create_patch: bool = False, **kwargs: Any) -> 'DiffIndex':
         """Creates diffs between two items being trees, trees and index or an
         index and the working tree. It will detect renames automatically.
 
@@ -89,11 +99,11 @@ class Diffable(object):
 
         :param paths:
             is a list of paths or a single path to limit the diff to.
-            It will only include at least one of the givne path or paths.
+            It will only include at least one of the given path or paths.
 
         :param create_patch:
             If True, the returned Diff contains a detailed patch that if applied
-            makes the self to other. Patches are somwhat costly as blobs have to be read
+            makes the self to other. Patches are somewhat costly as blobs have to be read
             and diffed.
 
         :param kwargs:
@@ -104,8 +114,8 @@ class Diffable(object):
 
         :note:
             On a bare repository, 'other' needs to be provided as Index or as
-            as Tree/Commit, or a git command error will occour"""
-        args = list()
+            as Tree/Commit, or a git command error will occur"""
+        args = []  # type: List[Union[str, Diffable, object]]
         args.append("--abbrev=40")        # we need full shas
         args.append("--full-index")       # get full index paths, not only filenames
 
@@ -114,6 +124,7 @@ class Diffable(object):
             args.append("-p")
         else:
             args.append("--raw")
+            args.append("-z")
 
         # in any way, assure we don't see colored output,
         # fixes https://github.com/gitpython-developers/GitPython/issues/172
@@ -122,6 +133,9 @@ class Diffable(object):
         if paths is not None and not isinstance(paths, (tuple, list)):
             paths = [paths]
 
+        if hasattr(self, 'repo'):  # else raise Error?
+            self.repo = self.repo  # type: 'Repo'
+
         diff_cmd = self.repo.git.diff
         if other is self.Index:
             args.insert(0, '--cached')
@@ -145,10 +159,10 @@ class Diffable(object):
         kwargs['as_process'] = True
         proc = diff_cmd(*self._process_diff_args(args), **kwargs)
 
-        diff_method = Diff._index_from_raw_format
-        if create_patch:
-            diff_method = Diff._index_from_patch_format
-        index = diff_method(self.repo, proc.stdout)
+        diff_method = (Diff._index_from_patch_format
+                       if create_patch
+                       else Diff._index_from_raw_format)
+        index = diff_method(self.repo, proc)
 
         proc.wait()
         return index
@@ -164,13 +178,14 @@ class DiffIndex(list):
     # A = Added
     # D = Deleted
     # R = Renamed
-    # M = modified
-    change_type = ("A", "D", "R", "M")
+    # M = Modified
+    # T = Changed in the type
+    change_type = ("A", "C", "D", "R", "M", "T")
 
-    def iter_change_type(self, change_type):
+    def iter_change_type(self, change_type: Lit_change_type) -> Iterator['Diff']:
         """
         :return:
-            iterator yieling Diff instances that match the given change_type
+            iterator yielding Diff instances that match the given change_type
 
         :param change_type:
             Member of DiffIndex.change_type, namely:
@@ -178,15 +193,21 @@ class DiffIndex(list):
             * 'A' for added paths
             * 'D' for deleted paths
             * 'R' for renamed paths
-            * 'M' for paths with modified data"""
+            * 'M' for paths with modified data
+            * 'T' for changed in the type paths
+         """
         if change_type not in self.change_type:
             raise ValueError("Invalid change type: %s" % change_type)
 
-        for diff in self:
-            if change_type == "A" and diff.new_file:
+        for diff in self:   # type: 'Diff'
+            if diff.change_type == change_type:
+                yield diff
+            elif change_type == "A" and diff.new_file:
                 yield diff
             elif change_type == "D" and diff.deleted_file:
                 yield diff
+            elif change_type == "C" and diff.copied_file:
+                yield diff
             elif change_type == "R" and diff.renamed:
                 yield diff
             elif change_type == "M" and diff.a_blob and diff.b_blob and diff.a_blob != diff.b_blob:
@@ -229,7 +250,7 @@ class Diff(object):
     # precompiled regex
     re_header = re.compile(br"""
                                 ^diff[ ]--git
-                                    [ ](?P<a_path_fallback>"?a/.+?"?)[ ](?P<b_path_fallback>"?b/.+?"?)\n
+                                    [ ](?P<a_path_fallback>"?[ab]/.+?"?)[ ](?P<b_path_fallback>"?[ab]/.+?"?)\n
                                 (?:^old[ ]mode[ ](?P<old_mode>\d+)\n
                                    ^new[ ]mode[ ](?P<new_mode>\d+)(?:\n|$))?
                                 (?:^similarity[ ]index[ ]\d+%\n
@@ -237,6 +258,9 @@ class Diff(object):
                                    ^rename[ ]to[ ](?P<rename_to>.*)(?:\n|$))?
                                 (?:^new[ ]file[ ]mode[ ](?P<new_file_mode>.+)(?:\n|$))?
                                 (?:^deleted[ ]file[ ]mode[ ](?P<deleted_file_mode>.+)(?:\n|$))?
+                                (?:^similarity[ ]index[ ]\d+%\n
+                                   ^copy[ ]from[ ].*\n
+                                   ^copy[ ]to[ ](?P<copied_file_name>.*)(?:\n|$))?
                                 (?:^index[ ](?P<a_blob_id>[0-9A-Fa-f]+)
                                     \.\.(?P<b_blob_id>[0-9A-Fa-f]+)[ ]?(?P<b_mode>.+)?(?:\n|$))?
                                 (?:^---[ ](?P<a_path>[^\t\n\r\f\v]*)[\t\r\f\v]*(?:\n|$))?
@@ -247,24 +271,33 @@ class Diff(object):
     NULL_BIN_SHA = b"\0" * 20
 
     __slots__ = ("a_blob", "b_blob", "a_mode", "b_mode", "a_rawpath", "b_rawpath",
-                 "new_file", "deleted_file", "raw_rename_from", "raw_rename_to", "diff")
+                 "new_file", "deleted_file", "copied_file", "raw_rename_from",
+                 "raw_rename_to", "diff", "change_type", "score")
 
-    def __init__(self, repo, a_rawpath, b_rawpath, a_blob_id, b_blob_id, a_mode,
-                 b_mode, new_file, deleted_file, raw_rename_from,
-                 raw_rename_to, diff):
+    def __init__(self, repo: 'Repo',
+                 a_rawpath: Optional[bytes], b_rawpath: Optional[bytes],
+                 a_blob_id: Union[str, bytes, None], b_blob_id: Union[str, bytes, None],
+                 a_mode: Union[bytes, str, None], b_mode: Union[bytes, str, None],
+                 new_file: bool, deleted_file: bool, copied_file: bool,
+                 raw_rename_from: Optional[bytes], raw_rename_to: Optional[bytes],
+                 diff: Union[str, bytes, None], change_type: Optional[str], score: Optional[int]) -> None:
 
-        self.a_mode = a_mode
-        self.b_mode = b_mode
-
-        assert a_rawpath is None or isinstance(a_rawpath, binary_type)
-        assert b_rawpath is None or isinstance(b_rawpath, binary_type)
+        assert a_rawpath is None or isinstance(a_rawpath, bytes)
+        assert b_rawpath is None or isinstance(b_rawpath, bytes)
         self.a_rawpath = a_rawpath
         self.b_rawpath = b_rawpath
 
-        if self.a_mode:
-            self.a_mode = mode_str_to_int(self.a_mode)
-        if self.b_mode:
-            self.b_mode = mode_str_to_int(self.b_mode)
+        self.a_mode = mode_str_to_int(a_mode) if a_mode else None
+        self.b_mode = mode_str_to_int(b_mode) if b_mode else None
+
+        # Determine whether this diff references a submodule, if it does then
+        # we need to overwrite "repo" to the corresponding submodule's repo instead
+        if repo and a_rawpath:
+            for submodule in repo.submodules:
+                if submodule.path == a_rawpath.decode(defenc, 'replace'):
+                    if submodule.module_exists():
+                        repo = submodule.module()
+                    break
 
         if a_blob_id is None or a_blob_id == self.NULL_HEX_SHA:
             self.a_blob = None
@@ -278,55 +311,60 @@ class Diff(object):
 
         self.new_file = new_file
         self.deleted_file = deleted_file
+        self.copied_file = copied_file
 
         # be clear and use None instead of empty strings
-        assert raw_rename_from is None or isinstance(raw_rename_from, binary_type)
-        assert raw_rename_to is None or isinstance(raw_rename_to, binary_type)
+        assert raw_rename_from is None or isinstance(raw_rename_from, bytes)
+        assert raw_rename_to is None or isinstance(raw_rename_to, bytes)
         self.raw_rename_from = raw_rename_from or None
         self.raw_rename_to = raw_rename_to or None
 
         self.diff = diff
+        self.change_type = change_type
+        self.score = score
 
-    def __eq__(self, other):
+    def __eq__(self, other: object) -> bool:
         for name in self.__slots__:
             if getattr(self, name) != getattr(other, name):
                 return False
         # END for each name
         return True
 
-    def __ne__(self, other):
+    def __ne__(self, other: object) -> bool:
         return not (self == other)
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash(tuple(getattr(self, n) for n in self.__slots__))
 
-    def __str__(self):
-        h = "%s"
+    def __str__(self) -> str:
+        h = "%s"  # type: str
         if self.a_blob:
             h %= self.a_blob.path
         elif self.b_blob:
             h %= self.b_blob.path
 
-        msg = ''
-        l = None    # temp line
-        ll = 0      # line length
+        msg = ''  # type: str
+        line = None          # temp line
+        line_length = 0      # line length
         for b, n in zip((self.a_blob, self.b_blob), ('lhs', 'rhs')):
             if b:
-                l = "\n%s: %o | %s" % (n, b.mode, b.hexsha)
+                line = "\n%s: %o | %s" % (n, b.mode, b.hexsha)
             else:
-                l = "\n%s: None" % n
+                line = "\n%s: None" % n
             # END if blob is not None
-            ll = max(len(l), ll)
-            msg += l
+            line_length = max(len(line), line_length)
+            msg += line
         # END for each blob
 
         # add headline
-        h += '\n' + '=' * ll
+        h += '\n' + '=' * line_length
 
         if self.deleted_file:
             msg += '\nfile deleted in rhs'
         if self.new_file:
             msg += '\nfile added in rhs'
+        if self.copied_file:
+            msg += '\nfile %r copied from %r' % (self.b_path, self.a_path)
         if self.rename_from:
             msg += '\nfile renamed from %r' % self.rename_from
         if self.rename_to:
@@ -334,53 +372,50 @@ class Diff(object):
         if self.diff:
             msg += '\n---'
             try:
-                msg += self.diff.decode(defenc)
+                msg += self.diff.decode(defenc) if isinstance(self.diff, bytes) else self.diff
             except UnicodeDecodeError:
                 msg += 'OMITTED BINARY DATA'
             # end handle encoding
             msg += '\n---'
         # END diff info
 
-        # Python2 sillyness: have to assure we convert our likely to be unicode object to a string with the
+        # Python2 silliness: have to assure we convert our likely to be unicode object to a string with the
         # right encoding. Otherwise it tries to convert it using ascii, which may fail ungracefully
         res = h + msg
-        if not PY3:
-            res = res.encode(defenc)
         # end
         return res
 
     @property
-    def a_path(self):
+    def a_path(self) -> Optional[str]:
         return self.a_rawpath.decode(defenc, 'replace') if self.a_rawpath else None
 
     @property
-    def b_path(self):
+    def b_path(self) -> Optional[str]:
         return self.b_rawpath.decode(defenc, 'replace') if self.b_rawpath else None
 
     @property
-    def rename_from(self):
+    def rename_from(self) -> Optional[str]:
         return self.raw_rename_from.decode(defenc, 'replace') if self.raw_rename_from else None
 
     @property
-    def rename_to(self):
+    def rename_to(self) -> Optional[str]:
         return self.raw_rename_to.decode(defenc, 'replace') if self.raw_rename_to else None
 
     @property
-    def renamed(self):
+    def renamed(self) -> bool:
         """:returns: True if the blob of our diff has been renamed
         :note: This property is deprecated, please use ``renamed_file`` instead.
         """
         return self.renamed_file
 
     @property
-    def renamed_file(self):
+    def renamed_file(self) -> bool:
         """:returns: True if the blob of our diff has been renamed
-        :note: This property is deprecated, please use ``renamed_file`` instead.
         """
         return self.rename_from != self.rename_to
 
     @classmethod
-    def _pick_best_path(cls, path_match, rename_match, path_fallback_match):
+    def _pick_best_path(cls, path_match: bytes, rename_match: bytes, path_fallback_match: bytes) -> Optional[bytes]:
         if path_match:
             return decode_path(path_match)
 
@@ -393,32 +428,41 @@ class Diff(object):
         return None
 
     @classmethod
-    def _index_from_patch_format(cls, repo, stream):
+    def _index_from_patch_format(cls, repo: 'Repo', proc: TBD) -> DiffIndex:
         """Create a new DiffIndex from the given text which must be in patch format
         :param repo: is the repository we are operating on - it is required
         :param stream: result of 'git diff' as a stream (supporting file protocol)
         :return: git.DiffIndex """
+
+        ## FIXME: Here SLURPING raw, need to re-phrase header-regexes linewise.
+        text_list = []  # type: List[bytes]
+        handle_process_output(proc, text_list.append, None, finalize_process, decode_streams=False)
+
         # for now, we have to bake the stream
-        text = stream.read()
+        text = b''.join(text_list)
         index = DiffIndex()
         previous_header = None
-        for header in cls.re_header.finditer(text):
+        header = None
+        a_path, b_path = None, None  # for mypy
+        a_mode, b_mode = None, None  # for mypy
+        for _header in cls.re_header.finditer(text):
             a_path_fallback, b_path_fallback, \
                 old_mode, new_mode, \
                 rename_from, rename_to, \
-                new_file_mode, deleted_file_mode, \
+                new_file_mode, deleted_file_mode, copied_file_name, \
                 a_blob_id, b_blob_id, b_mode, \
-                a_path, b_path = header.groups()
+                a_path, b_path = _header.groups()
 
-            new_file, deleted_file = bool(new_file_mode), bool(deleted_file_mode)
+            new_file, deleted_file, copied_file = \
+                bool(new_file_mode), bool(deleted_file_mode), bool(copied_file_name)
 
             a_path = cls._pick_best_path(a_path, rename_from, a_path_fallback)
             b_path = cls._pick_best_path(b_path, rename_to, b_path_fallback)
 
             # Our only means to find the actual text is to see what has not been matched by our regex,
-            # and then retro-actively assin it to our index
+            # and then retro-actively assign it to our index
             if previous_header is not None:
-                index[-1].diff = text[previous_header.end():header.start()]
+                index[-1].diff = text[previous_header.end():_header.start()]
             # end assign actual diff
 
             # Make sure the mode is set if the path is set. Otherwise the resulting blob is invalid
@@ -432,59 +476,81 @@ class Diff(object):
                               b_blob_id and b_blob_id.decode(defenc),
                               a_mode and a_mode.decode(defenc),
                               b_mode and b_mode.decode(defenc),
-                              new_file, deleted_file,
+                              new_file, deleted_file, copied_file,
                               rename_from,
                               rename_to,
-                              None))
+                              None, None, None))
 
-            previous_header = header
+            previous_header = _header
+            header = _header
         # end for each header we parse
-        if index:
+        if index and header:
             index[-1].diff = text[header.end():]
         # end assign last diff
 
         return index
 
-    @classmethod
-    def _index_from_raw_format(cls, repo, stream):
-        """Create a new DiffIndex from the given stream which must be in raw format.
-        :return: git.DiffIndex"""
-        # handles
-        # :100644 100644 687099101... 37c5e30c8... M    .gitignore
-        index = DiffIndex()
-        for line in stream.readlines():
-            line = line.decode(defenc)
-            if not line.startswith(":"):
-                continue
-            # END its not a valid diff line
-            meta, _, path = line[1:].partition('\t')
-            old_mode, new_mode, a_blob_id, b_blob_id, change_type = meta.split(None, 4)
+    @staticmethod
+    def _handle_diff_line(lines_bytes: bytes, repo: 'Repo', index: TBD) -> None:
+        lines = lines_bytes.decode(defenc)
+
+        for line in lines.split(':')[1:]:
+            meta, _, path = line.partition('\x00')
+            path = path.rstrip('\x00')
+            a_blob_id, b_blob_id = None, None  # Type: Optional[str]
+            old_mode, new_mode, a_blob_id, b_blob_id, _change_type = meta.split(None, 4)
+            # Change type can be R100
+            # R: status letter
+            # 100: score (in case of copy and rename)
+            change_type = _change_type[0]
+            score_str = ''.join(_change_type[1:])
+            score = int(score_str) if score_str.isdigit() else None
             path = path.strip()
             a_path = path.encode(defenc)
             b_path = path.encode(defenc)
             deleted_file = False
             new_file = False
+            copied_file = False
             rename_from = None
             rename_to = None
 
-            # NOTE: We cannot conclude from the existance of a blob to change type
+            # NOTE: We cannot conclude from the existence of a blob to change type
             # as diffs with the working do not have blobs yet
             if change_type == 'D':
-                b_blob_id = None
+                b_blob_id = None  # Optional[str]
                 deleted_file = True
             elif change_type == 'A':
                 a_blob_id = None
                 new_file = True
-            elif change_type[0] == 'R':     # parses RXXX, where XXX is a confidence value
-                a_path, b_path = path.split('\t', 1)
-                a_path = a_path.encode(defenc)
-                b_path = b_path.encode(defenc)
+            elif change_type == 'C':
+                copied_file = True
+                a_path_str, b_path_str = path.split('\x00', 1)
+                a_path = a_path_str.encode(defenc)
+                b_path = b_path_str.encode(defenc)
+            elif change_type == 'R':
+                a_path_str, b_path_str = path.split('\x00', 1)
+                a_path = a_path_str.encode(defenc)
+                b_path = b_path_str.encode(defenc)
                 rename_from, rename_to = a_path, b_path
+            elif change_type == 'T':
+                # Nothing to do
+                pass
             # END add/remove handling
 
             diff = Diff(repo, a_path, b_path, a_blob_id, b_blob_id, old_mode, new_mode,
-                        new_file, deleted_file, rename_from, rename_to, '')
+                        new_file, deleted_file, copied_file, rename_from, rename_to,
+                        '', change_type, score)
             index.append(diff)
-        # END for each line
+
+    @classmethod
+    def _index_from_raw_format(cls, repo: 'Repo', proc: TBD) -> DiffIndex:
+        """Create a new DiffIndex from the given stream which must be in raw format.
+        :return: git.DiffIndex"""
+        # handles
+        # :100644 100644 687099101... 37c5e30c8... M    .gitignore
+
+        index = DiffIndex()
+        handle_process_output(proc, lambda bytes: cls._handle_diff_line(
+            bytes, repo, index), None, finalize_process, decode_streams=False)
 
         return index
